# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jax import config

from jaxley.connect import connect

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")

import os
from math import pi

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"
import jax.numpy as jnp
import numpy as np
import pytest
from jaxley_mech.channels.l5pc import *

import jaxley as jx
from jaxley.channels import HH, K, Leak, Na
from jaxley.synapses import IonotropicSynapse


@pytest.mark.slow
@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
@pytest.mark.parametrize("file", ["morph_single_point_soma.swc", "morph.swc"])
def test_swc_cell(voltage_solver: str, file: str, SimpleMorphCell):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.2, delta_t=0.025, t_max=5.0
    )

    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "../swc_files", file)
    cell = SimpleMorphCell(fname, ncomp=2, max_branch_len=300.0)
    _ = cell.soma  # Only to test whether the `soma` group was created.
    cell.insert(HH())
    cell.branch(1).loc(0.0).record()
    cell.branch(1).loc(0.0).stimulate(current)

    voltages = jx.integrate(cell, delta_t=dt, voltage_solver=voltage_solver)

    if file == "morph_single_point_soma.swc":
        voltages_300724 = jnp.asarray(
            [
                [
                    -70.0,
                    -66.53085703217431,
                    -57.86539099487056,
                    -50.68402476185379,
                    -47.116280145195034,
                    -26.54674397910345,
                    25.86902654237883,
                    2.7545847203679648,
                    -23.49507300176727,
                    -46.176932023665,
                    -68.13277872699207,
                ]
            ]
        )
    elif file == "morph.swc":
        voltages_300724 = jnp.asarray(
            [
                [
                    -70.0,
                    -66.53085703215352,
                    -57.02081109294148,
                    -49.74061944051031,
                    -46.887867347343146,
                    -27.613724386581872,
                    25.092386748387153,
                    4.367651506995822,
                    -23.53096242684505,
                    -46.43371990417977,
                    -68.37487528619394,
                ]
            ]
        )
    else:
        raise NameError
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.slow
@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_swc_net(voltage_solver: str, SimpleMorphCell):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.2, delta_t=0.025, t_max=5.0
    )

    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "../swc_files/morph.swc")
    cell1 = SimpleMorphCell(fname, ncomp=2, max_branch_len=300.0)
    cell2 = SimpleMorphCell(fname, ncomp=2, max_branch_len=300.0)

    network = jx.Network([cell1, cell2])
    connect(
        network.cell(0).branch(0).loc(0.0),
        network.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )
    network.insert(HH())

    # first cell, 0-eth branch, 1-st compartment because loc=0.0 -> comp = ncomp-1 = 1
    radius_post = network[1, 0, 1].nodes["radius"].item()
    lenght_post = network[1, 0, 1].nodes["length"].item()
    area = 2 * pi * lenght_post * radius_post
    point_process_to_dist_factor = 100_000.0 / area
    network.set("IonotropicSynapse_gS", 0.5 / point_process_to_dist_factor)

    for cell_ind in range(2):
        network.cell(cell_ind).branch(1).loc(0.0).record()

    for stim_ind in range(2):
        network.cell(stim_ind).branch(1).loc(0.0).stimulate(current)

    voltages = jx.integrate(network, delta_t=dt, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703215763,
                -57.020811092945316,
                -49.74061944051095,
                -46.88786734735504,
                -27.613724386645945,
                25.092386748379845,
                4.3676515070377,
                -23.530962426808408,
                -46.433719904148994,
                -68.37487528619184,
            ],
            [
                -70.0,
                -66.52401203413784,
                -57.01048571919544,
                -49.723881897201146,
                -46.83983197382557,
                -27.35802551334217,
                25.122266645646633,
                4.1962717936840805,
                -23.669714625969657,
                -46.560355102122436,
                -68.4360112878163,
            ],
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


# This test will be skipped for now, due to weird quirk of the swc file, which has two
# different radii for the same xyz coordinates (L10364: 10362 -> 10549). This is handled
# by the differently by the two swc reader backends (graph seems to be the correct one,
# compared to NEURON).
@pytest.mark.skip
@pytest.mark.slow
@pytest.mark.parametrize("swc_backend", ["custom", "graph"])
def test_swc_morph(swc_backend, SimpleMorphCell):
    gt_apical = {}
    gt_soma = {}
    gt_axon = {}

    gt_apical["apical_NaTs2T_gNaTs2T"] = 0.026145
    gt_apical["apical_SKv3_1_gSKv3_1"] = 0.004226
    gt_apical["apical_M_gM"] = 0.000143

    gt_soma["somatic_NaTs2T_gNaTs2T"] = 0.983955
    gt_soma["somatic_SKv3_1_gSKv3_1"] = 0.303472
    gt_soma["somatic_SKE2_gSKE2"] = 0.008407
    gt_soma["somatic_CaPump_gamma"] = 0.000609
    gt_soma["somatic_CaPump_decay"] = 210.485291
    gt_soma["somatic_CaHVA_gCaHVA"] = 0.000994
    gt_soma["somatic_CaLVA_gCaLVA"] = 0.000333

    gt_axon["axonal_NaTaT_gNaTaT"] = 3.137968
    gt_axon["axonal_KPst_gKPst"] = 0.973538
    gt_axon["axonal_KTst_gKTst"] = 0.089259
    gt_axon["axonal_SKE2_gSKE2"] = 0.007104
    gt_axon["axonal_SKv3_1_gSKv3_1"] = 1.021945
    gt_axon["axonal_CaHVA_gCaHVA"] = 0.00099
    gt_axon["axonal_CaLVA_gCaLVA"] = 0.008752
    gt_axon["axonal_CaPump_gamma"] = 0.00291
    gt_axon["axonal_CaPump_decay"] = 287.19873

    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "../swc_files", "bbp_with_axon.swc")  # n120
    cell = SimpleMorphCell(fname, ncomp=2, swc_backend=swc_backend)

    # custom swc reader does not label the root branch that is added to the soma
    # while the graph swc reader does. This is accounted for here.
    cell.groups["soma"] = (
        cell.groups["soma"][2:] if swc_backend == "graph" else cell.groups["soma"]
    )
    apical_inds = cell.groups["apical"]

    ########## APICAL ##########
    cell.apical.set("capacitance", 2.0)
    cell.apical.insert(NaTs2T().change_name("apical_NaTs2T"))
    cell.apical.insert(SKv3_1().change_name("apical_SKv3_1"))
    cell.apical.insert(M().change_name("apical_M"))
    cell.apical.insert(H().change_name("apical_H"))

    for c in apical_inds:
        distance = cell.scope("global").comp(c).distance(cell.branch(1).loc(0.0))
        cond = (-0.8696 + 2.087 * np.exp(distance * 0.0031)) * 8e-5
        cell.scope("global").comp(c).set("apical_H_gH", cond)

    ########## SOMA ##########
    cell.soma.insert(NaTs2T().change_name("somatic_NaTs2T"))
    cell.soma.insert(SKv3_1().change_name("somatic_SKv3_1"))
    cell.soma.insert(SKE2().change_name("somatic_SKE2"))
    ca_dynamics = CaNernstReversal()
    ca_dynamics.channel_constants["T"] = 307.15
    cell.soma.insert(ca_dynamics)
    cell.soma.insert(CaPump().change_name("somatic_CaPump"))
    cell.soma.insert(CaHVA().change_name("somatic_CaHVA"))
    cell.soma.insert(CaLVA().change_name("somatic_CaLVA"))
    cell.soma.set("CaCon_i", 5e-05)
    cell.soma.set("CaCon_e", 2.0)

    ########## BASAL ##########
    cell.basal.insert(H().change_name("basal_H"))
    cell.basal.set("basal_H_gH", 8e-5)

    # ########## AXON ##########
    cell.insert(CaNernstReversal())
    cell.set("CaCon_i", 5e-05)
    cell.set("CaCon_e", 2.0)
    cell.axon.insert(NaTaT().change_name("axonal_NaTaT"))
    cell.axon.insert(KTst().change_name("axonal_KTst"))
    cell.axon.insert(CaPump().change_name("axonal_CaPump"))
    cell.axon.insert(SKE2().change_name("axonal_SKE2"))
    cell.axon.insert(CaHVA().change_name("axonal_CaHVA"))
    cell.axon.insert(KPst().change_name("axonal_KPst"))
    cell.axon.insert(SKv3_1().change_name("axonal_SKv3_1"))
    cell.axon.insert(CaLVA().change_name("axonal_CaLVA"))

    ########## WHOLE CELL  ##########
    cell.insert(Leak())
    cell.set("Leak_gLeak", 3e-05)
    cell.set("Leak_eLeak", -75.0)

    cell.set("axial_resistivity", 100.0)
    cell.set("eNa", 50.0)
    cell.set("eK", -85.0)
    cell.set("v", -65.0)

    for key in gt_apical.keys():
        cell.apical.set(key, gt_apical[key])

    for key in gt_soma.keys():
        cell.soma.set(key, gt_soma[key])

    for key in gt_axon.keys():
        cell.axon.set(key, gt_axon[key])

    dt = 0.025
    t_max = 100.0
    time_vec = np.arange(0, t_max + 2 * dt, dt)

    cell.delete_stimuli()
    cell.delete_recordings()

    i_delay = 10.0
    i_dur = 80.0
    i_amp = 3.0
    current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    cell.branch(1).comp(0).stimulate(current)
    cell.branch(1).comp(0).record()

    cell.set("v", -65.0)
    cell.init_states()

    voltages = jx.integrate(cell)

    voltages_250130 = jnp.asarray(
        [
            -65.0,
            -66.22422623,
            -67.23001452,
            -68.06298803,
            -68.75766951,
            -33.91317711,
            -55.24503749,
            -46.11452291,
            -42.18960646,
            -51.12861864,
            -43.65442616,
            -40.62727385,
            -49.56110473,
            -43.24030949,
            -36.71731271,
            -48.7405489,
            -42.98507829,
            -34.64282586,
            -48.24427898,
            -42.6412365,
            -34.70568206,
            -47.90643598,
            -42.15688181,
            -36.17711814,
            -47.65564274,
            -41.52265914,
            -38.1627371,
            -47.44680473,
            -40.70730741,
            -40.15298353,
            -47.25483146,
            -39.63994798,
            -41.96818737,
            -47.06569105,
            -38.17257448,
            -43.50053648,
            -46.87517934,
            -65.40488865,
            -69.96981343,
            -72.24384111,
            -73.46204372,
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::100] - voltages_250130))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
import pandas as pd


def remap_index_to_masked(
    index, nodes: pd.DataFrame, padded_cumsum_ncomp, ncomp_per_branch: jnp.ndarray
):
    """Convert actual index of the compartment to the index in the masked system.

    E.g. if `ncomps = [2, 4]`, then the index `3` would be mapped to `5` because the
    masked `ncomps` are `[4, 4]`. I.e.:

    original: [0, 1,           2, 3, 4, 5]
    masked:   [0, 1, (2) ,(3) ,4, 5, 6, 7]
    """
    cumsum_ncomp_per_branch = jnp.concatenate(
        [
            jnp.asarray([0]),
            jnp.cumsum(ncomp_per_branch),
        ]
    )
    branch_inds = nodes.loc[index, "global_branch_index"].to_numpy()
    remainders = index - cumsum_ncomp_per_branch[branch_inds]
    return padded_cumsum_ncomp[branch_inds] + remainders


def convert_to_csc(
    num_elements: int, row_ind: np.ndarray, col_ind: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Convert between two representations for sparse systems.

    This is needed because `jax.scipy.linalg.spsolve` requires the `(ind, indptr)`
    representation, but the `(row, col)` is more intuitive and easier to build.

    This function uses `np` instead of `jnp` because it only deals with indexing which
    can be dealt with only based on the branch structure (i.e. independent of any
    parameter values).

    Written by ChatGPT.
    """
    data_inds = np.arange(num_elements)
    # Step 1: Sort by (col_ind, row_ind)
    sorted_indices = np.lexsort((row_ind, col_ind))
    data_inds = data_inds[sorted_indices]
    row_ind = row_ind[sorted_indices]
    col_ind = col_ind[sorted_indices]

    # Step 2: Create indptr array
    n_cols = col_ind.max() + 1
    indptr = np.zeros(n_cols + 1, dtype=int)
    np.add.at(indptr, col_ind + 1, 1)
    np.cumsum(indptr, out=indptr)

    # Step 3: The row indices are already sorted
    indices = row_ind

    return data_inds, indices, indptr


def comp_edges_to_indices(
    comp_edges: pd.DataFrame,
) -> Tuple[int, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Generates sparse matrix indices from the table of node edges.

    This is only used for the `jax.sparse` voltage solver.

    Args:
        comp_edges: Dataframe with three columns (sink, source, type).

    Returns:
        n_nodes: The number of total nodes (including branchpoints).
        data_inds: The indices to reorder the data.
        indices and indptr: Indices passed to the sparse matrix solver.
    """
    # Build indices for diagonals.
    sources = np.asarray(comp_edges["source"].to_list())
    sinks = np.asarray(comp_edges["sink"].to_list())
    n_nodes = np.max(sinks) + 1 if len(sinks) > 0 else 1
    diagonal_inds = jnp.stack([jnp.arange(n_nodes), jnp.arange(n_nodes)])

    # Build indices for off-diagonals.
    off_diagonal_inds = jnp.stack([sources, sinks]).astype(int)

    # Concatenate indices of diagonals and off-diagonals.
    all_inds = jnp.concatenate([diagonal_inds, off_diagonal_inds], axis=1)

    # Cast (row, col) indices to the format required for the `jax` sparse solver.
    data_inds, indices, indptr = convert_to_csc(
        num_elements=all_inds.shape[1],
        row_ind=all_inds[0],
        col_ind=all_inds[1],
    )
    return n_nodes, data_inds, indices, indptr


class JaxleySolveIndexer:
    """Indexer for easy access to compartment indices given a branch index.

    Used only by the custom Jaxley solvers. This class has two purposes:

    1) It simplifies indexing. Indexing is difficult because every branch has a
    different number of compartments (in the solve, every branch within a level has
    the same number of compartments, but the number can still differ between levels).

    2) It stores several attributes such that we do not have to track all of them
    separately before they are used in `step()`.
    """

    def __init__(
        self,
        cumsum_ncomp: np.ndarray,
        ncomp_per_branch: np.ndarray,
        branchpoint_group_inds: Optional[np.ndarray] = None,
        children_in_level: Optional[np.ndarray] = None,
        parents_in_level: Optional[np.ndarray] = None,
        root_inds: Optional[np.ndarray] = None,
        remapped_node_indices: Optional[np.ndarray] = None,
    ):
        self.cumsum_ncomp = np.asarray(cumsum_ncomp)
        self.ncomp_per_branch = np.asarray(ncomp_per_branch)

        # Save items for easier access.
        self.branchpoint_group_inds = branchpoint_group_inds
        self.children_in_level = children_in_level
        self.parents_in_level = parents_in_level
        self.root_inds = root_inds
        self.remapped_node_indices = remapped_node_indices

    def first(self, branch_inds: np.ndarray) -> np.ndarray:
        """Return the indices of the first compartment of all `branch_inds`."""
        return self.cumsum_ncomp[branch_inds]

    def masked_last(self, branch_inds: np.ndarray) -> np.ndarray:
        """Return the indices of the last compartment of all `branch_inds`.

        Notably, this returns the index of the last compartment _with_ taking into
        account the masking structure which is needed to make all branches in a level
        have the same number of compartments.

        Example:
        ```
        parents = [-1, 0, 0, 1]
        ncomp_per_branch = [2, 2, 4, 4]
        masked_last([0, 1, 2, 3]) => [1, 5, 9, 13]
        # The important one is `masked_last(1) = 5`, because the second level has a
        # branch with 4 compartments.
        ```
        """
        return self.cumsum_ncomp[branch_inds + 1] - 1

    def last(self, branch_inds: np.ndarray) -> np.ndarray:
        """Return the indices of the last compartment of all `branch_inds`.

        Unlike `masked_last`, this returns the index of the last compartment that
        acutally has a value.

        Example:
        ```
        parents = [-1, 0, 0, 1]
        ncomp_per_branch = [2, 2, 4, 4]
        masked_last([0, 1, 2, 3]) => [1, 3, 9, 13]
        # The important one is `masked_last(1) = 3`, because the branch has only
        # 2 compartments.
        ```
        """
        return self.cumsum_ncomp[branch_inds] + self.ncomp_per_branch[branch_inds] - 1

    def branch(self, branch_inds: np.ndarray) -> np.ndarray:
        """Return indices of all compartments in all `branch_inds`."""
        start_inds = self.first(branch_inds)
        end_inds = self.masked_last(branch_inds) + 1
        return self._consecutive_indices(start_inds, end_inds)

    def lower(self, branch_inds: np.ndarray) -> np.ndarray:
        """Return indices of all lowers in all `branch_inds`.

        This is needed because the `lowers` array in the voltage solve is instantiated
        to have as many elements as the `diagonal`. In this method, we get rid of
        this additional element."""
        start_inds = self.first(branch_inds) + 1
        end_inds = self.masked_last(branch_inds) + 1
        return self._consecutive_indices(start_inds, end_inds)

    def upper(self, branch_inds: np.ndarray) -> np.ndarray:
        """Return indices of all uppers in all `branch_inds`.

        This is needed because the `uppers` array in the voltage solve is instantiated
        to have as many elements as the `diagonal`. In this method, we get rid of
        this additional element."""
        start_inds = self.first(branch_inds)
        end_inds = self.masked_last(branch_inds)
        return self._consecutive_indices(start_inds, end_inds)

    def _consecutive_indices(
        self, start_inds: np.ndarray, end_inds: np.ndarray
    ) -> np.ndarray:
        """Return array of all indices in [start, end], for every start, end.

        It also reshape the indices to `(nbranches, ncomp)`.

        E.g.:
        ```
        start_inds = [0, 6]
        end_inds = [3, 9]
        -->
        [[0, 1, 2], [6, 7, 8]]
        ```
        """
        n_inds = end_inds - start_inds
        assert np.all(n_inds[0] == n_inds), (
            "The indexer only supports indexing into branches with the same number "
            "of compartments."
        )
        if n_inds[0] > 0:
            repeated_starts = np.reshape(np.repeat(start_inds, n_inds), (-1, n_inds[0]))
            # For single compartment neurons there are no uppers or lowers, so `n_inds`
            # can be zero.
            return repeated_starts + np.arange(n_inds[0]).astype(int)
        else:
            return np.asarray([[]] * len(start_inds)).astype(int)

    def mask(self, indices: np.ndarray) -> np.ndarray:
        """Return the masked index given the global compartment index.

        The masked index is the one which occurs because all branches within a level
        must have the same number of compartments for the solve.
        """
        return self.remapped_node_indices[indices]



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jax import config

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")

import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"

from math import pi

import jax.numpy as jnp
import numpy as np
import pytest
from jaxley_mech.channels.l5pc import CaHVA

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect, fully_connect
from jaxley.pumps import CaFaradayConcentrationChange, CaNernstReversal
from jaxley.synapses import IonotropicSynapse, TestSynapse


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_compartment(voltage_solver, SimpleComp, SimpleBranch, SimpleCell, SimpleNet):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )
    tolerance = 1e-8

    voltages_081123 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703,
                -50.08343917,
                -20.4930498,
                32.6178327,
                2.57004029,
                -26.19517261,
                -50.50027566,
                -72.97019524,
                -75.69907171,
                -75.58874436,
            ]
        ]
    )

    # Test compartment.
    comp = SimpleComp()
    comp.insert(HH())
    comp.record()
    comp.stimulate(current)
    voltages = jx.integrate(comp, delta_t=dt, voltage_solver=voltage_solver)
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_081123))
    assert max_error <= tolerance, f"Compartment error is {max_error} > {tolerance}"

    # Test branch of a single compartment.
    branch = SimpleBranch(ncomp=1)
    branch.insert(HH())
    branch.record()
    branch.stimulate(current)
    voltages = jx.integrate(branch, delta_t=dt, voltage_solver=voltage_solver)
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_081123))
    assert max_error <= tolerance, f"Branch error is {max_error} > {tolerance}"

    # Test cell of a single compartment.
    cell = SimpleCell(1, 1)
    cell.insert(HH())
    cell.record()
    cell.stimulate(current)
    voltages = jx.integrate(cell, delta_t=dt, voltage_solver=voltage_solver)
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_081123))
    assert max_error <= tolerance, f"Cell error is {max_error} > {tolerance}"

    # Test net of a single compartment.
    net = SimpleNet(1, 1, 1)
    net.insert(HH())
    net.record()
    net.stimulate(current)
    voltages = jx.integrate(net, delta_t=dt, voltage_solver=voltage_solver)
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_081123))
    assert max_error <= tolerance, f"Network error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_branch(voltage_solver, SimpleBranch):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    branch = SimpleBranch(2)
    branch.insert(HH())
    branch.loc(0.0).record()
    branch.loc(0.0).stimulate(current)

    voltages = jx.integrate(branch, delta_t=dt, voltage_solver=voltage_solver)

    voltages_081123 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703,
                -56.84345977,
                -47.34841899,
                -32.63377325,
                30.99759112,
                8.03521283,
                -21.81119545,
                -45.8401624,
                -70.88960656,
                -75.66123707,
            ]
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_081123))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


def test_branch_fwd_euler_uneven_radiuses(SimpleBranch):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=2.0, delta_t=0.025, t_max=10.0
    )

    branch = SimpleBranch(8)
    branch.set("axial_resistivity", 500.0)

    rands1 = np.linspace(20, 300, 8)
    rands2 = np.linspace(1, 5, 8)
    branch.set("length", rands1)
    branch.set("radius", rands2)

    branch.insert(HH())
    branch.loc(1.0).stimulate(current)
    branch.loc(0.0).record()

    voltages = jx.integrate(branch, delta_t=dt, solver="fwd_euler")

    voltages_240920 = jnp.asarray(
        [
            -70.0,
            -64.319374,
            -61.61975,
            -56.971237,
            25.785686,
            -42.466354,
            -75.86178,
            -75.06558,
            -73.95041,
        ]
    )
    tolerance = 1e-5
    max_error = jnp.max(jnp.abs(voltages_240920 - voltages[0, ::50]))
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_cell(voltage_solver, SimpleCell):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    cell = SimpleCell(3, 2)
    cell.insert(HH())
    cell.branch(1).loc(0.0).record()
    cell.branch(1).loc(0.0).stimulate(current)

    voltages = jx.integrate(cell, delta_t=dt, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703217415,
                -61.30007495702876,
                -57.72869617299199,
                -57.907984804456134,
                -56.54785861252702,
                -54.63372123466727,
                -51.29073832791313,
                -41.84338223543283,
                12.629557592604355,
                16.375747329556404,
            ]
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_complex_cell(voltage_solver, SimpleBranch):
    """Test cell with a variety of channels, pumps, diffusion, and comps per branch."""
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )

    branch1 = SimpleBranch(ncomp=1)
    branch2 = SimpleBranch(ncomp=2)
    branch3 = SimpleBranch(ncomp=3)
    branch4 = SimpleBranch(ncomp=4)
    cell = jx.Cell([branch1, branch2, branch3, branch4], parents=[-1, 0, 0, 1])
    cell.set("axial_resistivity", 10_000.0)
    cell.insert(HH())

    # Calcium mechanisms.
    cell.branch(0).insert(CaHVA())
    cell.branch(0).set("CaHVA_gCaHVA", 0.0001)  # Larger than default to see impact on v
    cell.insert(CaFaradayConcentrationChange())
    cell.insert(CaNernstReversal())
    cell.diffuse("CaCon_i")
    # Large value to see the diffusion after a few milliseconds already.
    cell.set("axial_diffusion_CaCon_i", 180.2)

    cell.branch(1).comp(1).stimulate(current)
    cell.branch(0).comp(0).record("v")
    cell.branch(2).comp(2).record("v")
    cell.branch(3).comp(1).record("v")
    cell.branch(3).comp(3).record("v")
    cell.branch(0).comp(0).record("CaCon_i")
    cell.branch(2).comp(2).record("CaCon_i")
    cell.branch(3).comp(1).record("CaCon_i")
    cell.branch(3).comp(3).record("CaCon_i")

    recordings = jx.integrate(cell, delta_t=dt, voltage_solver=voltage_solver)
    voltages_240225 = jnp.asarray(
        [
            [-70.0, -53.80348826, 22.07746319, -47.29564472, -75.61534365],
            [-70.0, -61.74865308, 34.1167336, -28.89225642, -75.73529806],
            [-70.0, -52.99907533, 20.9735932, -47.83940886, -75.66904041],
            [-70.0, -60.86487129, 35.11018305, -33.89819309, -75.76933684],
        ]
    )
    cacon_240225 = (
        jnp.asarray(
            [
                [5.0, 5.04797687, 6.45337663, 12.1139712, 11.86289905],
                [5.0, 5.0063692, 5.02944485, 5.72251137, 7.49644484],
                [5.0, 5.00170413, 5.00819813, 5.20138788, 5.87906979],
                [5.0, 5.00020329, 5.00223143, 5.03325016, 5.27678607],
            ]
        )
        * 1e-5
    )
    max_error_ca = np.max(np.abs(recordings[4:, ::50] - cacon_240225))
    max_error_v = np.max(np.abs(recordings[:4, ::50] - voltages_240225))
    tolerance_ca = 1e-10
    tolerance_v = 1e-8
    assert max_error_ca <= tolerance_ca, f"Ca Error is {max_error_ca} > {tolerance_ca}"
    assert max_error_v <= tolerance_v, f"Voltage Error is {max_error_v} > {tolerance_v}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_net(voltage_solver, SimpleNet):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    net = SimpleNet(2, 3, 2)

    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )
    net.insert(HH())

    for cell_ind in range(2):
        net.cell(cell_ind).branch(1).loc(0.0).record()

    for stim_ind in range(2):
        net.cell(stim_ind).branch(1).loc(0.0).stimulate(current)

    area = 2 * pi * 10.0 * 1.0
    point_process_to_dist_factor = 100_000.0 / area
    net.IonotropicSynapse.set(
        "IonotropicSynapse_gS", 0.5 / point_process_to_dist_factor
    )
    voltages = jx.integrate(net, delta_t=dt, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703217415,
                -61.300074957028755,
                -57.728696172992,
                -57.90798480445608,
                -56.54785861252695,
                -54.633721234667185,
                -51.29073832791293,
                -41.843382235431946,
                12.629557592599513,
                16.37574732956111,
            ],
            [
                -70.0,
                -66.16039111452706,
                -60.50177192696074,
                -56.47792566140247,
                -55.905759661906906,
                -52.98761212851238,
                -46.3513320143946,
                -13.082888133400296,
                25.761263776404153,
                -4.627404050763336,
                -30.796093242248364,
            ],
        ]
    )

    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_complex_net(voltage_solver, SimpleNet):
    net = SimpleNet(7, 5, 4)
    net.insert(HH())

    _ = np.random.seed(0)
    pre = net.cell([0, 1, 2])
    post = net.cell([3, 4, 5])
    fully_connect(pre, post, IonotropicSynapse())
    fully_connect(pre, post, TestSynapse())

    pre = net.cell([3, 4, 5])
    post = net.cell(6)
    fully_connect(pre, post, IonotropicSynapse())
    fully_connect(pre, post, TestSynapse())

    area = 2 * pi * 10.0 * 1.0
    point_process_to_dist_factor = 100_000.0 / area
    net.set("IonotropicSynapse_gS", 0.44 / point_process_to_dist_factor)
    net.set("TestSynapse_gC", 0.62 / point_process_to_dist_factor)
    net.IonotropicSynapse.edge([0, 2, 4]).set(
        "IonotropicSynapse_gS", 0.32 / point_process_to_dist_factor
    )
    net.TestSynapse.edge([0, 3, 5]).set(
        "TestSynapse_gC", 0.24 / point_process_to_dist_factor
    )

    current = jx.step_current(
        i_delay=0.5, i_dur=0.5, i_amp=0.1, delta_t=0.025, t_max=10.0
    )
    for i in range(3):
        net.cell(i).branch(0).loc(0.0).stimulate(current)

    net.cell(6).branch(0).loc(0.0).record()

    voltages = jx.integrate(net, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -64.5478417407739,
                -61.37506280116741,
                -58.27492707181707,
                -51.53610508254835,
                31.739248535280243,
                -23.276048469250686,
                -73.58311542313007,
                -75.5489796956953,
                -74.69162422333675,
                -73.52874849932951,
            ]
        ]
    )

    max_error = np.max(np.abs(voltages[:, ::40] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Callable, Dict, List, Optional, Tuple, Union
from warnings import warn

import jax.numpy as jnp
import numpy as np
import pandas as pd

from jaxley.modules.base import Module
from jaxley.modules.branch import Branch
from jaxley.utils.cell_utils import (
    build_branchpoint_group_inds,
    compute_children_and_parents,
    compute_children_in_level,
    compute_children_indices,
    compute_levels,
    compute_morphology_indices_in_levels,
    compute_parents_in_level,
)
from jaxley.utils.misc_utils import cumsum_leading_zero, deprecated_kwargs
from jaxley.utils.solver_utils import (
    JaxleySolveIndexer,
    comp_edges_to_indices,
    remap_index_to_masked,
)


class Cell(Module):
    """Cell class.

    This class defines a single cell that can be simulated by itself or
    connected with synapses to build a network. A cell is made up of several branches
    and supports intricate cell morphologies.
    """

    cell_params: Dict = {}
    cell_states: Dict = {}

    def __init__(
        self,
        branches: Optional[Union[Branch, List[Branch]]] = None,
        parents: Optional[List[int]] = None,
        xyzr: Optional[List[np.ndarray]] = None,
    ):
        """Initialize a cell.

        Args:
            branches: A single branch or a list of branches that make up the cell.
                If a single branch is provided, then the branch is repeated `len(parents)`
                times to create the cell.
            parents: The parent branch index for each branch. The first branch has no
                parent and is therefore set to -1.
            xyzr: For every branch, the x, y, and z coordinates and the radius at the
                traced coordinates. Note that this is the full tracing (from SWC), not
                the stick representation coordinates.
        """
        super().__init__()
        assert (
            isinstance(branches, (Branch, List)) or branches is None
        ), "Only Branch or List[Branch] is allowed."
        if branches is not None:
            assert (
                parents is not None
            ), "If `branches` is not a list then you have to set `parents`."
        if isinstance(branches, List):
            assert len(parents) == len(
                branches
            ), "Ensure equally many parents, i.e. len(branches) == len(parents)."

        branches = Branch() if branches is None else branches
        parents = [-1] if parents is None else parents

        if isinstance(branches, Branch):
            branch_list = [branches for _ in range(len(parents))]
        else:
            branch_list = branches

        if xyzr is not None:
            assert len(xyzr) == len(parents)
            self.xyzr = xyzr
        else:
            # For every branch (`len(parents)`), we have a start and end point (`2`) and
            # a (x,y,z,r) coordinate for each of them (`4`).
            # Since `xyzr` is only inspected at `.vis()` and because it depends on the
            # (potentially learned) length of every compartment, we only populate
            # self.xyzr at `.vis()`.
            self.xyzr = [float("NaN") * np.zeros((2, 4)) for _ in range(len(parents))]

        self.total_nbranches = len(branch_list)
        self.nbranches_per_cell = [len(branch_list)]
        self.comb_parents = jnp.asarray(parents)
        self.comb_children = compute_children_indices(self.comb_parents)
        self._cumsum_nbranches = np.asarray([0, len(branch_list)])

        # Compartment structure. These arguments have to be rebuilt when `.set_ncomp()`
        # is run.
        self.ncomp_per_branch = np.asarray([branch.ncomp for branch in branch_list])
        self.ncomp = int(np.max(self.ncomp_per_branch))
        self.cumsum_ncomp = cumsum_leading_zero(self.ncomp_per_branch)
        self._internal_node_inds = np.arange(self.cumsum_ncomp[-1])

        # Build nodes. Has to be changed when `.set_ncomp()` is run.
        self.nodes = pd.concat([c.nodes for c in branch_list], ignore_index=True)
        self.nodes["global_comp_index"] = np.arange(self.cumsum_ncomp[-1])
        self.nodes["global_branch_index"] = np.repeat(
            np.arange(self.total_nbranches), self.ncomp_per_branch
        ).tolist()
        self.nodes["global_cell_index"] = np.repeat(0, self.cumsum_ncomp[-1]).tolist()
        self._update_local_indices()
        self._init_view()

        # Appending general parameters (radius, length, r_a, cm) and channel parameters,
        # as well as the states (v, and channel states).
        self._append_params_and_states(self.cell_params, self.cell_states)

        # Channels.
        self._gather_channels_from_constituents(branch_list)

        self.branch_edges = pd.DataFrame(
            dict(
                parent_branch_index=self.comb_parents[1:],
                child_branch_index=np.arange(1, self.total_nbranches),
            )
        )

        # For morphology indexing.
        self._par_inds, self._child_inds, self._child_belongs_to_branchpoint = (
            compute_children_and_parents(self.branch_edges)
        )

        self._initialize()

    def _init_morph_jaxley_spsolve(self):
        """Initialize morphology for the custom sparse solver.

        Running this function is only required for custom Jaxley solvers, i.e., for
        `voltage_solver={'jaxley.stone', 'jaxley.thomas'}`. However, because at
        `.__init__()` (when the function is run), we do not yet know which solver the
        user will use. Therefore, we always run this function at `.__init__()`.
        """
        children_and_parents = compute_morphology_indices_in_levels(
            len(self._par_inds),
            self._child_belongs_to_branchpoint,
            self._par_inds,
            self._child_inds,
        )
        branchpoint_group_inds = build_branchpoint_group_inds(
            len(self._par_inds),
            self._child_belongs_to_branchpoint,
            self.cumsum_ncomp[-1],
        )
        parents = self.comb_parents
        children_inds = children_and_parents["children"]
        parents_inds = children_and_parents["parents"]

        levels = compute_levels(parents)
        children_in_level = compute_children_in_level(levels, children_inds)
        parents_in_level = compute_parents_in_level(
            levels, self._par_inds, parents_inds
        )
        levels_and_ncomp = pd.DataFrame().from_dict(
            {
                "levels": levels,
                "ncomps": self.ncomp_per_branch,
            }
        )
        levels_and_ncomp["max_ncomp_in_level"] = levels_and_ncomp.groupby("levels")[
            "ncomps"
        ].transform("max")
        padded_cumsum_ncomp = cumsum_leading_zero(
            levels_and_ncomp["max_ncomp_in_level"].to_numpy()
        )

        # Generate mapping to deal with the masking which allows using the custom
        # sparse solver to deal with different ncomp per branch.
        remapped_node_indices = remap_index_to_masked(
            self._internal_node_inds,
            self.nodes,
            padded_cumsum_ncomp,
            self.ncomp_per_branch,
        )
        self._solve_indexer = JaxleySolveIndexer(
            cumsum_ncomp=padded_cumsum_ncomp,
            ncomp_per_branch=self.ncomp_per_branch,
            branchpoint_group_inds=branchpoint_group_inds,
            children_in_level=children_in_level,
            parents_in_level=parents_in_level,
            root_inds=np.asarray([0]),
            remapped_node_indices=remapped_node_indices,
        )

    def _init_morph_jax_spsolve(self):
        """For morphology indexing with the `jax.sparse` voltage volver.

        Explanation of `self._comp_eges['type']`:
        `type == 0`: compartment <--> compartment (within branch)
        `type == 1`: branchpoint --> parent-compartment
        `type == 2`: branchpoint --> child-compartment
        `type == 3`: parent-compartment --> branchpoint
        `type == 4`: child-compartment --> branchpoint

        Running this function is only required for generic sparse solvers, i.e., for
        `voltage_solver='jax.sparse'`.
        """

        # Edges between compartments within the branches.
        self._comp_edges = pd.concat(
            [
                pd.DataFrame()
                .from_dict(
                    {
                        "source": list(range(cumsum_ncomp, ncomp - 1 + cumsum_ncomp))
                        + list(range(1 + cumsum_ncomp, ncomp + cumsum_ncomp)),
                        "sink": list(range(1 + cumsum_ncomp, ncomp + cumsum_ncomp))
                        + list(range(cumsum_ncomp, ncomp - 1 + cumsum_ncomp)),
                    }
                )
                .astype(int)
                for ncomp, cumsum_ncomp in zip(self.ncomp_per_branch, self.cumsum_ncomp)
            ]
        )
        self._comp_edges["type"] = 0

        # Edges from branchpoints to compartments.
        branchpoint_to_parent_edges = pd.DataFrame().from_dict(
            {
                "source": np.arange(len(self._par_inds)) + self.cumsum_ncomp[-1],
                "sink": self.cumsum_ncomp[self._par_inds + 1] - 1,
                "type": 1,
            }
        )
        branchpoint_to_child_edges = pd.DataFrame().from_dict(
            {
                "source": self._child_belongs_to_branchpoint + self.cumsum_ncomp[-1],
                "sink": self.cumsum_ncomp[self._child_inds],
                "type": 2,
            }
        )
        self._comp_edges = pd.concat(
            [
                self._comp_edges,
                branchpoint_to_parent_edges,
                branchpoint_to_child_edges,
            ],
            ignore_index=True,
        )

        # Edges from compartments to branchpoints.
        parent_to_branchpoint_edges = branchpoint_to_parent_edges.rename(
            columns={"sink": "source", "source": "sink"}
        )
        parent_to_branchpoint_edges["type"] = 3
        child_to_branchpoint_edges = branchpoint_to_child_edges.rename(
            columns={"sink": "source", "source": "sink"}
        )
        child_to_branchpoint_edges["type"] = 4

        self._comp_edges = pd.concat(
            [
                self._comp_edges,
                parent_to_branchpoint_edges,
                child_to_branchpoint_edges,
            ],
            ignore_index=True,
        )

        n_nodes, data_inds, indices, indptr = comp_edges_to_indices(self._comp_edges)
        self._n_nodes = n_nodes
        self._data_inds = data_inds
        self._indices_jax_spsolve = indices
        self._indptr_jax_spsolve = indptr



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.optimize.optimizer import TypeOptimizer
from jaxley.optimize.transforms import ParamTransform



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax
import pytest

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


import jax.numpy as jnp
import numpy as np
from jax import jit, value_and_grad

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import fully_connect
from jaxley.synapses import IonotropicSynapse


def test_subclassing_groups_cell_api(SimpleCell):
    cell = SimpleCell(5, 4)

    cell.branch([0, 3, 4]).add_to_group("subtree")

    # The following lines are made possible by PR #324.
    cell.subtree.branch(0).set("radius", 0.1)
    cell.subtree.branch(0).comp("all").make_trainable("length")


def test_subclassing_groups_net_api(SimpleNet):
    net = SimpleNet(10, 2, 4)

    net.cell([0, 3, 5]).add_to_group("excitatory")

    # The following lines are made possible by PR #324.
    net.excitatory.cell(0).set("radius", 0.1)
    net.excitatory.cell(0).branch("all").make_trainable("length")


def test_subclassing_groups_net_set_equivalence(SimpleNet):
    """Test whether calling `.set` on subclasses group is same as on view."""
    net1 = SimpleNet(10, 2, 4)
    net2 = SimpleNet(10, 2, 4)

    net1.cell([0, 3, 5]).add_to_group("excitatory")

    # The following lines are made possible by PR #324.
    net1.excitatory.cell([0, 3]).branch(0).comp("all").set("radius", 0.14)
    net1.excitatory.cell([0, 5]).branch(1).comp("all").set("length", 0.16)
    net1.excitatory.cell("all").branch(1).comp(2).set("axial_resistivity", 1100.0)

    net2.cell([0, 3]).branch(0).comp("all").set("radius", 0.14)
    net2.cell([0, 5]).branch(1).comp("all").set("length", 0.16)
    net2.cell([0, 3, 5]).branch(1).comp(2).set("axial_resistivity", 1100.0)

    assert all(net1.nodes == net2.nodes)


def test_subclassing_groups_net_make_trainable_equivalence(SimpleNet):
    """Test whether calling `.maek_trainable` on subclasses group is same as on view."""
    net1 = SimpleNet(10, 2, 4)
    net2 = SimpleNet(10, 2, 4)

    net1.cell([0, 3, 5]).add_to_group("excitatory")

    # The following lines are made possible by PR #324.
    # The new behaviour needs changing of the scope to still conform here
    # TODO FROM #447: Rewrite this test / reconsider what behaviour is desired
    net1.excitatory.scope("global").cell([0, 3]).scope("local").branch(
        0
    ).make_trainable("radius")
    net1.excitatory.scope("global").cell([0, 5]).scope("local").branch(1).comp(
        "all"
    ).make_trainable("length")
    net1.excitatory.scope("global").cell("all").scope("local").branch(1).comp(
        2
    ).make_trainable("axial_resistivity")
    params1 = jnp.concatenate(jax.tree_util.tree_flatten(net1.get_parameters())[0])

    net2.cell([0, 3]).branch(0).make_trainable("radius")
    net2.cell([0, 5]).branch(1).comp("all").make_trainable("length")
    net2.cell([0, 3, 5]).branch(1).comp(2).make_trainable("axial_resistivity")
    params2 = jnp.concatenate(jax.tree_util.tree_flatten(net2.get_parameters())[0])
    assert jnp.array_equal(params1, params2)

    for inds1, inds2 in zip(
        net1.indices_set_by_trainables, net2.indices_set_by_trainables
    ):
        assert jnp.array_equal(inds1, inds2)


def test_fully_connect_groups_equivalence(SimpleNet):
    """Test whether groups can be used with `fully_connect`."""
    net1 = SimpleNet(10, 2, 4)
    net2 = SimpleNet(10, 2, 4)

    net1.cell([0, 3, 5]).add_to_group("layer1")
    net1.cell([6, 8]).add_to_group("layer2")

    pre = net1.layer1.cell("all")
    post = net1.layer2.cell("all")
    fully_connect(pre, post, IonotropicSynapse())

    pre = net2.cell([0, 3, 5])
    post = net2.cell([6, 8])
    fully_connect(pre, post, IonotropicSynapse())

    assert all(net1.edges == net2.edges)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.modules.base import Module
from jaxley.modules.branch import Branch
from jaxley.modules.cell import Cell
from jaxley.modules.compartment import Compartment
from jaxley.modules.network import Network



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional, Tuple, Union
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from numpy import ndarray
from scipy.spatial import ConvexHull

from jaxley.utils.cell_utils import v_interp


def plot_graph(
    xyzr: ndarray,
    dims: Tuple[int] = (0, 1),
    color: str = "k",
    ax: Optional[Axes] = None,
    type: str = "line",
    **kwargs,
) -> Axes:
    """Plot morphology.

    Args:
        xyzr: The coordinates of the morphology.
        dims: Which dimensions to plot. 1=x, 2=y, 3=z coordinate. Must be a tuple of
            two or three of them.
        color: The color for all branches.
        ax: The matplotlib axis to plot on.
        type: Either `line` or `scatter`.
        kwargs: The plot kwargs for plt.plot or plt.scatter.
    """

    if ax is None:
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_subplot(111) if len(dims) < 3 else plt.axes(projection="3d")

    for coords_of_branch in xyzr:
        points = coords_of_branch[:, dims].T

        if "line" in type.lower():
            _ = ax.plot(*points, color=color, **kwargs)
        elif "scatter" in type.lower():
            _ = ax.scatter(*points, color=color, **kwargs)
        else:
            raise NotImplementedError

    return ax


def extract_outline(points: ndarray) -> ndarray:
    """Get the outline of a 2D/3D shape.

    Extracts the subset of points which form the convex hull, i.e. the outline of
    the input points.

    Args:
        points: An array of points / corrdinates.

    Returns:
        An array of points which form the convex hull.
    """
    hull = ConvexHull(points)
    hull_points = points[hull.vertices]
    return hull_points


def compute_rotation_matrix(axis: ndarray, angle: float) -> ndarray:
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by the given angle.

    Can be used to rotate a coordinate vector by multiplying it with the rotation
    matrix.

    Args:
        axis: The axis of rotation.
        angle: The angle of rotation in radians.

    Returns:
        A 3x3 rotation matrix.
    """
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(angle / 2.0)
    b, c, d = -axis * np.sin(angle / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array(
        [
            [aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc],
        ]
    )


def create_cone_frustum_mesh(
    length: float,
    radius_bottom: float,
    radius_top: float,
    bottom_dome: bool = False,
    top_dome: bool = False,
    resolution: int = 100,
) -> ndarray:
    """Generates mesh points for a cone frustum, with optional domes at either end.

    This is used to render the traced morphology in 3D (and to project it to 2D)
    as part of `plot_morph`. Sections between two traced coordinates with two
    different radii can be represented by a cone frustum. Additionally, the ends
    of the frustum can be capped with hemispheres to ensure that two neighbouring
    frustums are connected smoothly (like ball joints).

    Args:
        length: The length of the frustum.
        radius_bottom: The radius of the bottom of the frustum.
        radius_top: The radius of the top of the frustum.
        bottom_dome: If True, a dome is added to the bottom of the frustum.
            The dome is a hemisphere with radius `radius_bottom`.
        top_dome: If True, a dome is added to the top of the frustum.
            The dome is a hemisphere with radius `radius_top`.
        resolution: defines the resolution of the mesh.
            If too low (typically <10), can result in errors.
            Useful too have a simpler mesh for plotting.

    Returns:
        An array of mesh points.
    """

    t = np.linspace(0, 2 * np.pi, resolution)

    # Determine the total height including domes
    total_height = length
    total_height += radius_bottom if bottom_dome else 0
    total_height += radius_top if top_dome else 0

    z = np.linspace(0, total_height, resolution)
    t_grid, z_coords = np.meshgrid(t, z)

    # Initialize arrays
    x_coords = np.zeros_like(t_grid)
    y_coords = np.zeros_like(t_grid)
    r_coords = np.zeros_like(t_grid)

    # Bottom hemisphere
    if bottom_dome:
        dome_mask = z_coords < radius_bottom
        arg = 1 - z_coords[dome_mask] / radius_bottom
        arg[np.isclose(arg, 1, atol=1e-6, rtol=1e-6)] = 1
        arg[np.isclose(arg, -1, atol=1e-6, rtol=1e-6)] = -1
        phi = np.arccos(1 - z_coords[dome_mask] / radius_bottom)
        r_coords[dome_mask] = radius_bottom * np.sin(phi)
        z_coords[dome_mask] = z_coords[dome_mask]

    # Frustum
    frustum_start = radius_bottom if bottom_dome else 0
    frustum_end = total_height - (radius_top if top_dome else 0)
    frustum_mask = (z_coords >= frustum_start) & (z_coords <= frustum_end)
    z_frustum = z_coords[frustum_mask] - frustum_start
    r_coords[frustum_mask] = radius_bottom + (radius_top - radius_bottom) * (
        z_frustum / length
    )

    # Top hemisphere
    if top_dome:
        dome_mask = z_coords > (total_height - radius_top)
        arg = (z_coords[dome_mask] - (total_height - radius_top)) / radius_top
        arg[np.isclose(arg, 1, atol=1e-6, rtol=1e-6)] = 1
        arg[np.isclose(arg, -1, atol=1e-6, rtol=1e-6)] = -1
        phi = np.arccos(arg)
        r_coords[dome_mask] = radius_top * np.sin(phi)

    x_coords = r_coords * np.cos(t_grid)
    y_coords = r_coords * np.sin(t_grid)

    return np.stack([x_coords, y_coords, z_coords])


def create_cylinder_mesh(
    length: float, radius: float, resolution: int = 100
) -> ndarray:
    """Generates mesh points for a cylinder.

    This is used to render cylindrical compartments in 3D (and to project it to 2D)
    as part of `plot_comps`.

    Args:
        length: The length of the cylinder.
        radius: The radius of the cylinder.
        resolution: defines the resolution of the mesh.
            If too low (typically <10), can result in errors.
            Useful too have a simpler mesh for plotting.

    Returns:
        An array of mesh points.
    """
    # Define cylinder
    t = np.linspace(0, 2 * np.pi, resolution)
    z_coords = np.linspace(-length / 2, length / 2, resolution)
    t_grid, z_coords = np.meshgrid(t, z_coords)

    x_coords = radius * np.cos(t_grid)
    y_coords = radius * np.sin(t_grid)
    return np.stack([x_coords, y_coords, z_coords])


def create_sphere_mesh(radius: float, resolution: int = 100) -> np.ndarray:
    """Generates mesh points for a sphere.

    This is used to render spherical compartments in 3D (and to project it to 2D)
    as part of `plot_comps`.

    Args:
        radius: The radius of the sphere.
        resolution: defines the resolution of the mesh.
            If too low (typically <10), can result in errors.
            Useful too have a simpler mesh for plotting.

    Returns:
        An array of mesh points.
    """
    phi = np.linspace(0, np.pi, resolution)
    theta = np.linspace(0, 2 * np.pi, resolution)

    # Create a 2D meshgrid for phi and theta
    phi_coords, theta_coords = np.meshgrid(phi, theta)

    # Convert spherical coordinates to Cartesian coordinates
    x_coords = radius * np.sin(phi_coords) * np.cos(theta_coords)
    y_coords = radius * np.sin(phi_coords) * np.sin(theta_coords)
    z_coords = radius * np.cos(phi_coords)

    return np.stack([x_coords, y_coords, z_coords])


def plot_mesh(
    mesh_points: ndarray,
    orientation: ndarray,
    center: ndarray,
    dims: Tuple[int],
    ax: Axes = None,
    **kwargs,
) -> Axes:
    """Plot the 2D projection of a volume mesh on a cardinal plane.

    Project the projection of a cylinder that is oriented in 3D space.
    - Create cylinder mesh
    - rotate cylinder mesh to orient it lengthwise along a given orientation vector.
    - move its center
    - project onto plane
    - compute outline of projected mesh.
    - fill area inside the outline

    Args:
        mesh_points: coordinates of the xyz mesh that define the volume
        orientation: orientation vector. The cylinder will be oriented along this vector.
        center: The x,y,z coordinates of the center of the cylinder.
        dims: The dimensions to plot / to project the cylinder onto,
        i.e. [0,1] xy-plane or [0,1,2] for 3D.
        ax: The matplotlib axis to plot on.

    Returns:
        Plot of the cylinder projection.
    """
    if ax is None:
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_subplot(111) if len(dims) < 3 else plt.axes(projection="3d")

    # Normalize axis vector
    orientation = np.array(orientation)
    orientation = orientation / np.linalg.norm(orientation)

    # Create a rotation matrix to align the cylinder with the given axis
    z_axis = np.array([0, 0, 1])
    rotation_axis = np.cross(z_axis, orientation)
    rotation_angle = np.arccos(np.dot(z_axis, orientation))

    if np.allclose(rotation_axis, 0):
        rotation_matrix = np.eye(3)
    else:
        rotation_matrix = compute_rotation_matrix(rotation_axis, rotation_angle)

    # Rotate mesh
    x_mesh, y_mesh, z_mesh = mesh_points
    rotated_mesh_points = np.dot(
        rotation_matrix,
        np.array([x_mesh.flatten(), y_mesh.flatten(), z_mesh.flatten()]),
    )
    rotated_mesh_points = rotated_mesh_points.reshape(3, -1)

    # project onto plane and move
    rotated_mesh_points = rotated_mesh_points[dims]
    rotated_mesh_points += np.array(center)[dims, np.newaxis]

    if len(dims) < 3:
        # get outline of cylinder mesh
        mesh_outline = extract_outline(rotated_mesh_points.T).T
        ax.fill(*mesh_outline.reshape(mesh_outline.shape[0], -1), **kwargs)
    else:
        # plot 3d mesh
        ax.plot_surface(*rotated_mesh_points.reshape(*mesh_points.shape), **kwargs)
    return ax


def plot_comps(
    module_or_view: Union["jx.Module", "jx.View"],
    dims: Tuple[int] = (0, 1),
    color: str = "k",
    ax: Optional[Axes] = None,
    true_comp_length: bool = True,
    resolution: int = 100,
    **kwargs,
) -> Axes:
    """Plot compartmentalized neural morphology.

    Plots the projection of the cylindrical compartments.

    Args:
        module_or_view: The module or view to plot.
        dims: The dimensions to plot / to project the cylinder onto,
            i.e. [0,1] xy-plane or [0,1,2] for 3D.
        color: The color for all compartments
        ax: The matplotlib axis to plot on.
        true_comp_length: If True, the length of the compartment is used, i.e. the
            length of the traced neurite. This means for zig-zagging neurites the
            cylinders will be longer than the straight-line distance between the
            start and end point of the neurite. This can lead to overlapping and
            miss-aligned cylinders. Setting this False will use the straight-line
            distance instead for nicer plots.
        resolution: defines the resolution of the mesh.
            If too low (typically <10), can result in errors.
            Useful too have a simpler mesh for plotting.
        kwargs: The plot kwargs for plt.fill.

    Returns:
        Plot of the compartmentalized morphology.
    """
    if ax is None:
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_subplot(111) if len(dims) < 3 else plt.axes(projection="3d")

    assert not np.any(
        np.isnan(module_or_view.xyzr[0][:, :3])
    ), "missing xyz coordinates."
    if "x" not in module_or_view.nodes.columns:
        module_or_view.compute_compartment_centers()

    for idx, xyzr in zip(module_or_view._branches_in_view, module_or_view.xyzr):
        locs = xyzr[:, :3]
        if locs.shape[0] == 1:  # assume spherical comp
            radius = xyzr[:, -1]
            center = xyzr[0, :3]
            if len(dims) == 3:
                xyz = create_sphere_mesh(radius, resolution)
                ax = plot_mesh(
                    xyz,
                    np.array([0, 0, 1]),
                    center,
                    np.array(dims),
                    ax,
                    color=color,
                    **kwargs,
                )
            else:
                ax.add_artist(plt.Circle(locs[0, dims], radius, color=color))
        else:
            lens = np.sqrt(np.nansum(np.diff(locs, axis=0) ** 2, axis=1))
            lens = np.cumsum([0] + lens.tolist())
            comp_ends = v_interp(
                np.linspace(0, lens[-1], module_or_view.ncomp + 1), lens, locs
            ).T
            axes = np.diff(comp_ends, axis=0)
            cylinder_lens = np.sqrt(np.sum(axes**2, axis=1))

            branch_df = module_or_view.nodes[
                module_or_view.nodes["global_branch_index"] == idx
            ]
            for l, axis, (i, comp) in zip(cylinder_lens, axes, branch_df.iterrows()):
                center = comp[["x", "y", "z"]]
                radius = comp["radius"]
                length = comp["length"] if true_comp_length else l
                xyz = create_cylinder_mesh(length, radius, resolution)
                ax = plot_mesh(
                    xyz,
                    axis,
                    center,
                    np.array(dims),
                    ax,
                    color=color,
                    **kwargs,
                )
    return ax


def plot_morph(
    module_or_view: Union["jx.Module", "jx.View"],
    dims: Tuple[int] = (0, 1),
    color: str = "k",
    ax: Optional[Axes] = None,
    resolution: int = 100,
    **kwargs,
) -> Axes:
    """Plot the detailed morphology.

    Plots the traced morphology it was traced. That means at every point that was
    traced a disc of radius `r` is plotted. The outline of the discs are then
    connected to form the morphology. This means every trace segement can be
    represented by a cone frustum. To prevent breaks in the morphology, each
    segement is connected with a ball joint.

    Args:
        module_or_view: The module or view to plot.
        dims: The dimensions to plot / to project the cylinder onto,
            i.e. [0,1] xy-plane or [0,1,2] for 3D.
        color: The color for all branches
        ax: The matplotlib axis to plot on.
        kwargs: The plot kwargs for plt.fill.

        resolution: defines the resolution of the mesh.
            If too low (typically <10), can result in errors.
            Useful too have a simpler mesh for plotting.

    Returns:
        Plot of the detailed morphology."""
    if ax is None:
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_subplot(111) if len(dims) < 3 else plt.axes(projection="3d")
    if len(dims) == 3:
        warn(
            "rendering large morphologies in 3D can take a while. Consider projecting to 2D instead."
        )

    assert not np.any(
        np.isnan(module_or_view.xyzr[0][:, :3])
    ), "missing xyz coordinates."

    for xyzr in module_or_view.xyzr:
        if len(xyzr) > 1:
            for xyzr1, xyzr2 in zip(xyzr[1:, :], xyzr[:-1, :]):
                dxyz = xyzr2[:3] - xyzr1[:3]
                length = np.sqrt(np.sum(dxyz**2))
                points = create_cone_frustum_mesh(
                    length,
                    xyzr1[-1],
                    xyzr2[-1],
                    bottom_dome=True,
                    top_dome=True,
                    resolution=resolution,
                )
                plot_mesh(
                    points,
                    dxyz,
                    xyzr1[:3],
                    np.array(dims),
                    color=color,
                    ax=ax,
                    **kwargs,
                )
        else:
            points = create_cone_frustum_mesh(
                0,
                xyzr[:, -1],
                xyzr[:, -1],
                bottom_dome=True,
                top_dome=True,
                resolution=resolution,
            )
            plot_mesh(
                points,
                np.ones(3),
                xyzr[0, :3],
                dims=np.array(dims),
                color=color,
                ax=ax,
                **kwargs,
            )

    return ax



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from morph_tool import convert


def run_conversion():
    convert("morph_tmp.swc", "morph.swc", sanitize=True, single_point_soma=True)


if __name__ == "__main__":
    run_conversion()



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")
from typing import Dict, Optional

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH, CaL, CaT, Channel, K, Km, Leak, Na
from jaxley.solver_gate import exponential_euler, save_exp, solve_inf_gate_exponential


class CaPump(Channel):
    """Calcium dynamics tracking inside calcium concentration, modeled after Destexhe et al. 1994."""

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {
            f"{self._name}_gamma": 0.05,  # Fraction of free calcium (not buffered)
            f"{self._name}_decay": 80,  # Rate of removal of calcium in ms
            f"{self._name}_depth": 0.1,  # Depth of shell in um
            f"{self._name}_minCai": 1e-4,  # Minimum intracellular calcium concentration in mM
        }
        self.channel_states = {
            f"CaCon_i": 5e-05,  # Initial internal calcium concentration in mM
        }
        self.current_name = f"i_Ca"
        self.META = {
            "reference": "Modified from Destexhe et al., 1994",
            "mechanism": "Calcium dynamics",
        }

    def update_states(self, u, dt, voltages, params):
        """Update internal calcium concentration based on calcium current and decay."""
        prefix = self._name
        ica = u["i_Ca"] / 1_000.0
        cai = u["CaCon_i"]
        gamma = params[f"{prefix}_gamma"]
        decay = params[f"{prefix}_decay"]
        depth = params[f"{prefix}_depth"]
        minCai = params[f"{prefix}_minCai"]

        FARADAY = 96485  # Coulombs per mole

        # Calculate the contribution of calcium currents to cai change
        drive_channel = -10_000.0 * ica * gamma / (2 * FARADAY * depth)

        cai_tau = decay
        cai_inf = minCai + decay * drive_channel
        new_cai = exponential_euler(cai, dt, cai_inf, cai_tau)

        return {f"CaCon_i": new_cai}

    def compute_current(self, u, voltages, params):
        """This dynamics model does not directly contribute to the membrane current."""
        return 0

    def init_state(self, states, voltages, params, delta_t):
        """Initialize the state at fixed point of gate dynamics."""
        return {}


class CaNernstReversal(Channel):
    """Compute Calcium reversal from inner and outer concentration of calcium."""

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_constants = {
            "F": 96485.3329,  # C/mol (Faraday's constant)
            "T": 279.45,  # Kelvin (temperature)
            "R": 8.314,  # J/(mol K) (gas constant)
        }
        self.channel_params = {}
        self.channel_states = {"eCa": 0.0, "CaCon_i": 5e-05, "CaCon_e": 2.0}
        self.current_name = f"i_Ca"

    def update_states(self, u, dt, voltages, params):
        """Update internal calcium concentration based on calcium current and decay."""
        R, T, F = (
            self.channel_constants["R"],
            self.channel_constants["T"],
            self.channel_constants["F"],
        )
        Cai = u["CaCon_i"]
        Cao = u["CaCon_e"]
        C = R * T / (2 * F) * 1000  # mV
        vCa = C * jnp.log(Cao / Cai)
        return {"eCa": vCa, "CaCon_i": Cai, "CaCon_e": Cao}

    def compute_current(self, u, voltages, params):
        """This dynamics model does not directly contribute to the membrane current."""
        return 0

    def init_state(self, states, voltages, params, delta_t):
        """Initialize the state at fixed point of gate dynamics."""
        return {}


def test_channel_set_name():
    # default name is the class name
    assert Na().name == "Na"

    # channel name can be set in the constructor
    na = Na(name="NaPospischil")
    assert na.name == "NaPospischil"
    assert "NaPospischil_gNa" in na.channel_params.keys()
    assert "eNa" in na.channel_params.keys()
    assert "NaPospischil_h" in na.channel_states.keys()
    assert "NaPospischil_m" in na.channel_states.keys()
    assert "NaPospischil_vt" not in na.channel_params.keys()
    assert "vt" in na.channel_params.keys()

    # channel name can not be changed directly
    k = K()
    with pytest.raises(AttributeError):
        k.name = "KPospischil"
    assert "KPospischil_gNa" not in k.channel_params.keys()
    assert "eNa" not in k.channel_params.keys()
    assert "KPospischil_h" not in k.channel_states.keys()
    assert "KPospischil_m" not in k.channel_states.keys()


def test_channel_change_name():
    # channel name can be changed with change_name method
    # (and only this way after initialization)
    na = Na().change_name("NaPospischil")
    assert na.name == "NaPospischil"
    assert "NaPospischil_gNa" in na.channel_params.keys()
    assert "eNa" in na.channel_params.keys()
    assert "NaPospischil_h" in na.channel_states.keys()
    assert "NaPospischil_m" in na.channel_states.keys()
    assert "NaPospischil_vt" not in na.channel_params.keys()
    assert "vt" in na.channel_params.keys()


def test_integration_with_renamed_channels():
    neuron_hh = HH().change_name("NeuronHH")
    standard_hh = HH()

    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=4)

    branch.loc(0.0).insert(standard_hh)
    branch.insert(neuron_hh)

    branch.loc(1.0).record()
    v = jx.integrate(branch, t_max=1.0)

    # Test if voltage is `NaN` which happens when channels get mixed up.
    assert np.invert(np.any(np.isnan(v)))


@pytest.mark.slow
def test_init_states(SimpleCell):
    """Functional test for `init_states()`.

    Checks whether, if everything is initialized in its steady state, the voltage
    after 10ms is almost exactly the same as after 0ms.
    """
    cell = SimpleCell(2, 4)
    cell.branch(0).loc(0.0).record()

    cell.branch(0).insert(Na())
    cell.branch(1).insert(K())
    cell.branch(1).loc(0.0).insert(Km())
    cell.branch(0).loc(1.0).insert(CaT())
    cell.insert(CaL())
    cell.insert(Leak())

    cell.insert(HH())

    cell.set("v", -62.0)  # At -70.0 there is a rebound spike.
    cell.init_states()
    v = jx.integrate(cell, t_max=20.0)

    last_voltage = v[0, -1]
    cell.set("v", last_voltage)
    cell.init_states()

    v = jx.integrate(cell, t_max=10.0)
    assert np.abs(v[0, 0] - v[0, -1]) < 0.02


class KCA11(Channel):
    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        prefix = self._name
        self.channel_params = {
            f"{prefix}_q10_ch": 3,
            f"{prefix}_q10_ch0": 22,
            "celsius": 22,
        }
        self.channel_states = {f"{prefix}_m": 0.02, "CaCon_i": 1e-4}
        self.current_name = f"i_K"

    def update_states(
        self,
        states: Dict[str, jnp.ndarray],
        dt,
        v,
        params: Dict[str, jnp.ndarray],
    ):
        """Update state."""
        prefix = self._name
        m = states[f"{prefix}_m"]
        q10 = params[f"{prefix}_q10_ch"] ** (
            (params["celsius"] - params[f"{prefix}_q10_ch0"]) / 10
        )
        cai = states["CaCon_i"]
        new_m = solve_inf_gate_exponential(m, dt, *self.m_gate(v, cai, q10))
        return {f"{prefix}_m": new_m}

    def compute_current(
        self, states: Dict[str, jnp.ndarray], v, params: Dict[str, jnp.ndarray]
    ):
        """Return current."""
        prefix = self._name
        m = states[f"{prefix}_m"]
        g = 0.03 * m * 1000  # mS/cm^2
        return g * (v + 80.0)

    def init_state(self, states, v, params, dt):
        """Initialize the state such at fixed point of gate dynamics."""
        prefix = self._name
        q10 = params[f"{prefix}_q10_ch"] ** (
            (params["celsius"] - params[f"{prefix}_q10_ch0"]) / 10
        )
        cai = states["CaCon_i"]
        m_inf, _ = self.m_gate(v, cai, q10)
        return {f"{prefix}_m": m_inf}

    @staticmethod
    def m_gate(v, cai, q10):
        cai = cai * 1e3
        v_half = -66 + 137 * save_exp(-0.3044 * cai) + 30.24 * save_exp(-0.04141 * cai)
        alpha = 25.0

        beta = 0.075 / save_exp((v - v_half) / 10)
        m_inf = alpha / (alpha + beta)
        tau_m = 1.0 * q10
        return m_inf, tau_m


def test_init_states_complex_channel(SimpleCell):
    """Test for `init_states()` with a more complicated channel model.

    The channel model used for this test uses the `states` in `init_state` and it also
    uses `q10`. The model inserts the channel only is some branches. This test follows
    an issue I had with Jaxley in v0.2.0 (fixed in v0.2.1).
    """
    ## Create cell
    cell = SimpleCell(3, 1)

    # CA channels.
    cell.branch([0, 1]).insert(CaNernstReversal())
    cell.branch([0, 1]).insert(CaPump())
    cell.branch([0, 1]).insert(KCA11())

    cell.init_states()

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    cell.branch(2).comp(0).stimulate(current)
    cell.branch(2).comp(0).record()
    voltages = jx.integrate(cell)
    assert np.invert(np.any(np.isnan(voltages))), "NaN voltage found"


def test_multiple_channel_currents(SimpleCell):
    """Test whether all channels can"""

    class User(Channel):
        """The channel which uses currents of Dummy1 and Dummy2 to update its states."""

        def __init__(self, name: Optional[str] = None):
            self.current_is_in_mA_per_cm2 = True
            super().__init__(name)
            self.channel_params = {}
            self.channel_states = {"cumulative": 0.0}
            self.current_name = f"i_User"

        def update_states(self, states, dt, v, params):
            state = states["cumulative"]
            state += states["i_Dummy"] * 0.001
            return {"cumulative": state}

        def compute_current(self, states, v, params):
            return 0.01 * jnp.ones_like(v)

    class Dummy1(Channel):
        def __init__(self, name: Optional[str] = None):
            self.current_is_in_mA_per_cm2 = True
            super().__init__(name)
            self.channel_params = {}
            self.channel_states = {}
            self.current_name = f"i_Dummy"

        def update_states(self, states, dt, v, params):
            return {}

        def compute_current(self, states, v, params):
            return 0.01 * jnp.ones_like(v)

    class Dummy2(Channel):
        def __init__(self, name: Optional[str] = None):
            self.current_is_in_mA_per_cm2 = True
            super().__init__(name)
            self.channel_params = {}
            self.channel_states = {}
            self.current_name = f"i_Dummy"

        def update_states(self, states, dt, v, params):
            return {}

        def compute_current(self, states, v, params):
            return 0.01 * jnp.ones_like(v)

    dt = 0.025  # ms
    t_max = 5.0  # ms
    cell = SimpleCell(1, 1)
    cell.branch(0).loc(0.0).stimulate(
        jx.step_current(i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0)
    )

    cell.insert(User())
    cell.insert(Dummy1())
    cell.insert(Dummy2())
    cell.branch(0).loc(0.0).record("cumulative")

    s = jx.integrate(cell, delta_t=dt)

    num_channels = 2
    target = (t_max // dt + 2) * 0.001 * 0.01 * num_channels
    assert np.abs(target - s[0, -1]) < 1e-8


def test_delete_channel(SimpleBranch):
    # test complete removal of a channel from a module
    branch1 = SimpleBranch(ncomp=3)
    branch1.comp(0).insert(K())
    branch1.delete(K())

    branch2 = SimpleBranch(ncomp=3)
    branch2.comp(0).insert(K())
    branch2.comp(0).delete(K())

    branch3 = SimpleBranch(ncomp=3)
    branch3.insert(K())
    branch3.delete(K())

    def channel_present(view, channel, partial=False):
        states_and_params = list(channel.channel_states.keys()) + list(
            channel.channel_params.keys()
        )
        # none of the states or params should be in nodes
        cols = view.nodes.columns.to_list()
        channel_cols = [
            col
            for col in cols
            if col.startswith(channel._name) and col != channel._name
        ]
        diff = set(channel_cols).difference(set(states_and_params))
        has_params_or_states = len(diff) > 0
        has_channel_col = channel._name in view.nodes.columns
        has_channel = channel._name in [c._name for c in view.channels]
        has_mem_current = channel.current_name in view.membrane_current_names
        if partial:
            all_nans = (
                not view.nodes[channel_cols].isna().all().all()
                & ~view.nodes[channel._name].all()
            )
            return has_channel or has_mem_current or all_nans
        return has_params_or_states or has_channel_col or has_channel or has_mem_current

    for branch in [branch1, branch2, branch3]:
        assert len(branch.channels) == 0
        assert not channel_present(branch, K())

    # test correct channels are removed only in the viewed part of the module
    branch4 = SimpleBranch(ncomp=3)
    branch4.insert(HH())
    branch4.comp(0).insert(K())
    branch4.comp([1, 2]).insert(Leak())

    branch4.comp(1).delete(Leak())
    # assert K in comp 0 and Leak still present in branch
    assert channel_present(branch4.comp(0), K())
    assert channel_present(branch4.comp(2), Leak(), partial=True)
    assert not channel_present(branch4.comp(1), Leak(), partial=True)
    assert channel_present(branch4, Leak())

    branch4.comp(2).delete(Leak())
    # assert no more Leak
    assert not channel_present(branch4, Leak())



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

VERSION = (0, 7, 0)

__version__ = ".".join(map(str, VERSION))



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import time
import warnings

import pytest

import jaxley as jx

pytest.skip(allow_module_level=True)


def test_module_retrieval(SimpleNet):
    t0 = time.time()
    comp = jx.Compartment()
    branch = jx.Branch([comp] * 4)
    cell = jx.Cell([branch] * 4, [-1, 0, 0, 1])
    net = jx.Network([cell] * 2)
    t1 = time.time()

    net = SimpleNet(2, 4, 4, force_init=False)
    t2 = time.time()

    assert ((t2 - t1) - (t1 - t0)) / (
        t1 - t0
    ) < 0.1, f"Fixture is slower than manual init."

    net = SimpleNet(2, 4, 4, force_init=False)
    t3 = time.time()
    assert (
        t1 - t0 > t2 - t1 > t3 - t2
    ), f"T_get: from pre-existing fixture {t3 - t2}, from fixture: {(t2 - t1)}, manual: {(t1 - t0)}"


def test_direct_submodule_retrieval(SimpleBranch):
    t1 = time.time()
    branch = SimpleBranch(2, 3, force_init=False)
    t2 = time.time()
    branch = SimpleBranch(4, 3, force_init=False)
    t3 = time.time()
    assert (
        t2 - t1 > t3 - t2
    ), f"T_get: from pre-existing fixture {t3 - t2}, from fixture: {(t2 - t1)}"


def test_recursive_submodule_retrieval(SimpleNet):
    t1 = time.time()
    net = SimpleNet(3, 4, 3, force_init=False)
    t2 = time.time()
    net = SimpleNet(3, 4, 3, force_init=False)
    t3 = time.time()
    assert (
        t2 - t1 > t3 - t2
    ), f"T_get: from pre-existing fixture {t3 - t2}, from fixture: {(t2 - t1)}"


def test_module_reinit(SimpleComp):
    t0 = time.time()
    comp = jx.Compartment()
    t1 = time.time()

    comp = SimpleComp(force_init=False)

    t2 = time.time()
    comp = SimpleComp(force_init=False)
    t3 = time.time()
    net = SimpleComp(force_init=True)
    t4 = time.time()

    msg = f"T_get: reinit {t4 - t3}, from fixture: {(t3 - t2)}, manual: {(t1 - t0)}"
    assert t1 - t0 > t4 - t3 or abs(((t1 - t0) - (t4 - t3)) / (t1 - t0)) < 0.3, msg
    assert t4 - t3 > t3 - t2, msg
